/*
 * Hexxagon board game.
 * Copyright (C) 2001 Erik Jonsson.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * Email erik@nesqi.homeip.net
 * 
 */

#include "hexxagonboard.h"
#include "hexxagonmove.h"

#include <sys/timeb.h>
#include <iostream.h>
#include <stdlib.h>

void HexxagonMoveList::addMove(HexxagonMove &move)
{
 	moves[nr_moves] = move;
	nr_moves++;
}

int compare(const void *mp1, const void *mp2)
{
	return ((HexxagonMove*)mp2)->score - ((HexxagonMove*)mp1)->score;
}

void HexxagonMoveList::sortList()
{
	qsort(moves, nr_moves, sizeof(moves[0]), &compare);
}

int getTime()
{
	struct timeb tb;
	
	ftime(&tb);
	return (tb.time * 1000) + tb.millitm;
}

HexxagonMove *HexxagonMoveList::getMove(int i)
{
	return &moves[i];
}

int HexxagonMoveList::getNrMoves()
{
	return nr_moves;
}

int alphaBeta(HexxagonBoard &board, int level, int alpha, int beta)
{
	int best = -INFINITY;

	if(!level)
		return -board.evaluate();

	HexxagonMoveList *moves = board.generateMoveList();

	if(!moves) // The game is over !
		return -board.evaluate();


	for(int i = 0; (i < moves->getNrMoves()) && (best < beta); i++)
	{
		if(best > alpha)
			alpha = best;

		HexxagonBoard newboard = HexxagonBoard(board);
		newboard.applyMove(*(moves->getMove(i)));
		int value = -alphaBeta(newboard, level - 1, -beta, -alpha);

		if(value > best)
			best = value;
	}     

	delete moves;

	if(!moves->getNrMoves() == 0)
		return board.evaluate();

	return best;
}

void HexxagonMoveList::scoreAllMoves(HexxagonBoard board, int depth, void (*callback)(), int maxtime)
{
	int alpha = -INFINITY;
	int beta = INFINITY;
	int best = -INFINITY;

	int t = getTime();

	for(int i = 1; (i <= depth) && (getTime() - t <= maxtime); i++)
	{
		for(int j = 0; (j < getNrMoves()) && (getTime() - t <= maxtime); j++)
		{
			if(best > alpha)
				alpha = best;

			HexxagonBoard newboard = HexxagonBoard(board);
			newboard.applyMove(*getMove(j));
			int value = -alphaBeta(newboard, i, -beta, -alpha);
			
			getMove(j)->score = value;

			if(value > best)
				best = value;
		}
		
		sortList();
	}

	cout << "Move computed in " << ((double)getTime() - t) / (double)1000 << " seconds.\n";
}

HexxagonMove *HexxagonMoveList::getBestMove()
{
	srandom(getTime());

	if(nr_moves)
	{
		int max = getMove(0)->score;
		int nr = 1;
	
		while(nr <= nr_moves && getMove(nr)->score == max)
			nr++;

		return getMove(random() % nr);
	}
	
	return NULL;
}




